/******************************************************************************
 *{@C
 *      Copyright:      2005-2006 Paul Obermeier (obermeier@poSoft.de)
 *
 *                      See the file "Tcl3D_License.txt" for information on
 *                      usage and redistribution of this file, and for a
 *                      DISCLAIMER OF ALL WARRANTIES.
 *
 *      Module:         Tcl3D
 *      Filename:       tcl3dArrays.i
 *
 *      Author:         Paul Obermeier
 *
 *      Description:    SWIG file defining utility functions for accessing
 *                      elements of a C array.
 *                      Slightly enhanced version of the carrays.i of the SWIG
 *                      standard library.
 *
 *****************************************************************************/

/* -----------------------------------------------------------------------------
 * %baseTypeVector(TYPE,NAME)
 *
 * Generates functions for creating and accessing elements of a C array
 * (as pointers).  Creates the following functions:
 *
 *        TYPE *new_NAME(int nelements)
 *        void delete_NAME(TYPE *);
 *        TYPE NAME_getitem(TYPE *, int index);
 *        void NAME_setitem(TYPE *, int index, TYPE value);
 * 
 * ----------------------------------------------------------------------------- */

%define %baseTypeVector(TYPE,NAME)
%{
static TYPE *new_##NAME(int nelements) { %}
#ifdef __cplusplus
%{  return new TYPE[nelements]; %}
#else
%{  return (TYPE *) calloc(nelements,sizeof(TYPE)); %}
#endif
%{}

static void delete_##NAME(TYPE *ary) { %}
#ifdef __cplusplus
%{  delete [] ary; %}
#else
%{  free(ary); %}
#endif
%{}

static TYPE NAME##_getitem(TYPE *ary, int index) {
    return ary[index];
}

static void NAME##_setitem(TYPE *ary, int index, TYPE value) {
    ary[index] = value;
}

static void NAME##_setarray(TYPE *ary, TYPE value, int startIndex, int len) {
    int i;
    int endIndex = startIndex + len;
    for (i=startIndex; i<endIndex; i++) {
        ary[i] = value;
    }
}

static void NAME##_addarray(TYPE *ary, double value, int startIndex, int len) {
    int i;
    int endIndex = startIndex + len;
    for (i=startIndex; i<endIndex; i++) {
        ary[i] += (TYPE) value;
    }
}

static void NAME##_mularray(TYPE *ary, double value, int startIndex, int len) {
    int i;
    int endIndex = startIndex + len;
    for (i=startIndex; i<endIndex; i++) {
        ary[i] *= (TYPE) value;
    }
}

static TYPE *NAME##_ind(TYPE *ary, int incr) {
    return (ary + incr);
}
static TYPE *NAME##_convert(void *ary) {
    return (TYPE *)ary;
}
%}

TYPE *new_##NAME(int nelements);
void delete_##NAME(TYPE *ary);
TYPE NAME##_getitem(TYPE *ary, int index);
void NAME##_setitem(TYPE *ary, int index, TYPE value);
void NAME##_setarray(TYPE *ary, TYPE value, int startIndex, int len);
void NAME##_addarray(TYPE *ary, double value, int startIndex, int len);
void NAME##_mularray(TYPE *ary, double value, int startIndex, int len);
TYPE *NAME##_ind(TYPE *ary, int incr);
TYPE *NAME##_convert(void *ary);

%enddef


%define %complexTypeVector(TYPE,NAME)
%{
static TYPE *new_##NAME(int nelements) { %}
#ifdef __cplusplus
%{  return new TYPE[nelements]; %}
#else
%{  return (TYPE *) calloc(nelements,sizeof(TYPE)); %}
#endif
%{}

static void delete_##NAME(TYPE *ary) { %}
#ifdef __cplusplus
%{  delete [] ary; %}
#else
%{  free(ary); %}
#endif
%{}

static TYPE NAME##_getitem(TYPE *ary, int index) {
    return ary[index];
}
static void NAME##_setitem(TYPE *ary, int index, TYPE value) {
    ary[index] = value;
}


static TYPE *NAME##_ind(TYPE *ary, int incr) {
    return (ary + incr);
}
static TYPE *NAME##_convert(void *ary) {
    return (TYPE *)ary;
}
%}

TYPE *new_##NAME(int nelements);
void delete_##NAME(TYPE *ary);
TYPE NAME##_getitem(TYPE *ary, int index);
void NAME##_setitem(TYPE *ary, int index, TYPE value);
TYPE *NAME##_ind(TYPE *ary, int incr);
TYPE *NAME##_convert(void *ary);

%enddef

// Generate array functions (new, delete, getitem, setitem, convert)
// for the following base types.
%baseTypeVector(char,char)
%baseTypeVector(short,short)
%baseTypeVector(int,int)
%baseTypeVector(float,float)
%baseTypeVector(double,double)

// Generate array functions (new, delete, getitem, setitem, convert)
// for the following OpenGL types.
// These associations must correspond with the settings in gl.h

%baseTypeVector(unsigned int,GLenum)
%baseTypeVector(unsigned char,GLboolean)
%baseTypeVector(unsigned int,GLbitfield)
%baseTypeVector(signed char,GLbyte)
%baseTypeVector(short,GLshort)
%baseTypeVector(int,GLint)
%baseTypeVector(int,GLsizei)
%baseTypeVector(unsigned char,GLubyte)
%baseTypeVector(unsigned short,GLushort)
%baseTypeVector(unsigned int,GLuint)
%baseTypeVector(float,GLfloat)
%baseTypeVector(float,GLclampf)
%baseTypeVector(double,GLdouble)
%baseTypeVector(double,GLclampd)
%baseTypeVector(char,GLchar)
%baseTypeVector(char,GLcharARB)
